# -*- coding: utf-8 -*-

# PLEASE DO NOT EDIT THIS FILE, IT IS GENERATED AND WILL BE OVERWRITTEN:
# https://github.com/ccxt/ccxt/blob/master/CONTRIBUTING.md#how-to-contribute-code

import ccxt.async_support
from ccxt.async_support.base.ws.cache import ArrayCache, ArrayCacheBySymbolById, ArrayCacheBySymbolBySide, ArrayCacheByTimestamp
from ccxt.base.types import Any, Balances, Int, Market, Order, OrderBook, Position, Str, Strings, Ticker, Tickers, Trade
from ccxt.async_support.base.ws.client import Client
from typing import List


class xt(ccxt.async_support.xt):

    def describe(self) -> Any:
        return self.deep_extend(super(xt, self).describe(), {
            'has': {
                'ws': True,
                'watchOHLCV': True,
                'watchOrderBook': True,
                'watchTicker': True,
                'watchTickers': True,
                'watchTrades': True,
                'watchTradesForSymbols': False,
                'watchBalance': True,
                'watchOrders': True,
                'watchMyTrades': True,
                'watchPositions': True,
            },
            'urls': {
                'api': {
                    'ws': {
                        'spot': 'wss://stream.xt.com',
                        'contract': 'wss://fstream.xt.com/ws',
                    },
                },
            },
            'options': {
                'tradesLimit': 1000,
                'ordersLimit': 1000,
                'OHLCVLimit': 1000,
                'watchTicker': {
                    'method': 'ticker',  # agg_ticker(contract only)
                },
                'watchTickers': {
                    'method': 'tickers',  # agg_tickers(contract only)
                },
                'watchPositions': {
                    'type': 'swap',
                    'fetchPositionsSnapshot': True,
                    'awaitPositionsSnapshot': True,
                },
            },
            'streaming': {
                'keepAlive': 20000,
                'ping': self.ping,
            },
            'token': None,
        })

    async def get_listen_key(self, isContract: bool):
        """
 @ignore
        required for private endpoints
        :param str isContract: True for contract trades

        https://doc.xt.com/#websocket_privategetToken
        https://doc.xt.com/#futures_user_websocket_v2base

        :returns str: listen key / access token
        """
        self.check_required_credentials()
        tradeType = 'contract' if isContract else 'spot'
        url = self.urls['api']['ws'][tradeType]
        if not isContract:
            url = url + '/private'
        client = self.client(url)
        token = self.safe_string(client.subscriptions, 'token')
        if token is None:
            if isContract:
                response = await self.privateLinearGetFutureUserV1UserListenKey()
                #
                #    {
                #        returnCode: '0',
                #        msgInfo: 'success',
                #        error: null,
                #        result: '3BC1D71D6CF96DA3458FC35B05B633351684511731128'
                #    }
                #
                client.subscriptions['token'] = self.safe_string(response, 'result')
            else:
                response = await self.privateSpotPostWsToken()
                #
                #    {
                #        "rc": 0,
                #        "mc": "SUCCESS",
                #        "ma": [],
                #        "result": {
                #            "token": "eyJhbqGciOiJSUzI1NiJ9.eyJhY2NvdW50SWQiOiIyMTQ2Mjg1MzIyNTU5Iiwic3ViIjoibGh4dDRfMDAwMUBzbmFwbWFpbC5jYyIsInNjb3BlIjoiYXV0aCIsImlzcyI6Inh0LmNvbSIsImxhc3RBdXRoVGltZSI6MTY2MzgxMzY5MDk1NSwic2lnblR5cGUiOiJBSyIsInVzZXJOYW1lIjoibGh4dDRfMDAwMUBzbmFwbWFpbC5jYyIsImV4cCI6MTY2NjQwNTY5MCwiZGV2aWNlIjoidW5rbm93biIsInVzZXJJZCI6MjE0NjI4NTMyMjU1OX0.h3zJlJBQrK2x1HvUxsKivnn6PlSrSDXXXJ7WqHAYSrN2CG5XPTKc4zKnTVoYFbg6fTS0u1fT8wH7wXqcLWXX71vm0YuP8PCvdPAkUIq4-HyzltbPr5uDYd0UByx0FPQtq1exvsQGe7evXQuDXx3SEJXxEqUbq_DNlXPTq_JyScI",
                #            "refreshToken": "eyJhbGciOiqJSUzI1NiJ9.eyJhY2NvdW50SWQiOiIyMTQ2Mjg1MzIyNTU5Iiwic3ViIjoibGh4dDRfMDAwMUBzbmFwbWFpbC5jYyIsInNjb3BlIjoicmVmcmVzaCIsImlzcyI6Inh0LmNvbSIsImxhc3RBdXRoVGltZSI6MTY2MzgxMzY5MDk1NSwic2lnblR5cGUiOiJBSyIsInVzZXJOYW1lIjoibGh4dDRfMDAwMUBzbmFwbWFpbC5jYyIsImV4cCI6MTY2NjQwNTY5MCwiZGV2aWNlIjoidW5rbm93biIsInVzZXJJZCI6MjE0NjI4NTMyMjU1OX0.Fs3YVm5YrEOzzYOSQYETSmt9iwxUHBovh2u73liv1hLUec683WGfktA_s28gMk4NCpZKFeQWFii623FvdfNoteXR0v1yZ2519uNvNndtuZICDdv3BQ4wzW1wIHZa1skxFfqvsDnGdXpjqu9UFSbtHwxprxeYfnxChNk4ssei430"
                #        }
                #    }
                #
                result = self.safe_dict(response, 'result')
                client.subscriptions['token'] = self.safe_string(result, 'accessToken')
        return client.subscriptions['token']

    def get_cache_index(self, orderbook, cache):
        # return the first index of the cache that can be applied to the orderbook or -1 if not possible
        nonce = self.safe_integer(orderbook, 'nonce')
        firstDelta = self.safe_value(cache, 0)
        firstDeltaNonce = self.safe_integer_2(firstDelta, 'i', 'u')
        if nonce < firstDeltaNonce - 1:
            return -1
        for i in range(0, len(cache)):
            delta = cache[i]
            deltaNonce = self.safe_integer_2(delta, 'i', 'u')
            if deltaNonce >= nonce:
                return i
        return len(cache)

    def handle_delta(self, orderbook, delta):
        orderbook['nonce'] = self.safe_integer_2(delta, 'i', 'u')
        obAsks = self.safe_list(delta, 'a', [])
        obBids = self.safe_list(delta, 'b', [])
        bids = orderbook['bids']
        asks = orderbook['asks']
        for i in range(0, len(obBids)):
            bid = obBids[i]
            price = self.safe_number(bid, 0)
            quantity = self.safe_number(bid, 1)
            bids.store(price, quantity)
        for i in range(0, len(obAsks)):
            ask = obAsks[i]
            price = self.safe_number(ask, 0)
            quantity = self.safe_number(ask, 1)
            asks.store(price, quantity)
        # self.handleBidAsks(storedBids, bids)
        # self.handleBidAsks(storedAsks, asks)

    async def subscribe(self, name: str, access: str, methodName: str, market: Market = None, symbols: List[str] = None, params={}):
        """
 @ignore
        Connects to a websocket channel

        https://doc.xt.com/#websocket_privaterequestFormat
        https://doc.xt.com/#futures_market_websocket_v2base

        :param str name: name of the channel
        :param str access: public or private
        :param str methodName: the name of the CCXT class method
        :param dict [market]: CCXT market
        :param str[] [symbols]: unified market symbols
        :param dict params: extra parameters specific to the xt api
        :returns dict: data from the websocket stream
        """
        privateAccess = access == 'private'
        type = None
        type, params = self.handle_market_type_and_params(methodName, market, params)
        isContract = (type != 'spot')
        subscribe = {
            'method': 'SUBSCRIBE' if isContract else 'subscribe',
            'id': self.number_to_string(self.milliseconds()) + name,  # call back ID
        }
        if privateAccess:
            if not isContract:
                subscribe['params'] = [name]
                subscribe['listenKey'] = await self.get_listen_key(isContract)
            else:
                listenKey = await self.get_listen_key(isContract)
                param = name + '@' + listenKey
                subscribe['params'] = [param]
        else:
            subscribe['params'] = [name]
        tradeType = 'contract' if isContract else 'spot'
        messageHash = name + '::' + tradeType
        if symbols is not None:
            messageHash = messageHash + '::' + ','.join(symbols)
        request = self.extend(subscribe, params)
        tail = access
        if isContract:
            tail = 'user' if privateAccess else 'market'
        url = self.urls['api']['ws'][tradeType] + '/' + tail
        return await self.watch(url, messageHash, request, messageHash)

    async def watch_ticker(self, symbol: str, params={}) -> Ticker:
        """
        watches a price ticker, a statistical calculation with the information calculated over the past 24 hours for a specific market

        https://doc.xt.com/#websocket_publictickerRealTime
        https://doc.xt.com/#futures_market_websocket_v2tickerRealTime
        https://doc.xt.com/#futures_market_websocket_v2aggTickerRealTime

        :param str symbol: unified symbol of the market to fetch the ticker for
        :param dict params: extra parameters specific to the xt api endpoint
        :param str [params.method]: 'agg_ticker'(contract only) or 'ticker', default = 'ticker' - the endpoint that will be streamed
        :returns dict: a `ticker structure <https://docs.ccxt.com/en/latest/manual.html#ticker-structure>`
        """
        await self.load_markets()
        market = self.market(symbol)
        options = self.safe_dict(self.options, 'watchTicker')
        defaultMethod = self.safe_string(options, 'method', 'ticker')
        method = self.safe_string(params, 'method', defaultMethod)
        name = method + '@' + market['id']
        return await self.subscribe(name, 'public', 'watchTicker', market, None, params)

    async def watch_tickers(self, symbols: Strings = None, params={}) -> Tickers:
        """
        watches a price ticker, a statistical calculation with the information calculated over the past 24 hours for a specific market

        https://doc.xt.com/#websocket_publicallTicker
        https://doc.xt.com/#futures_market_websocket_v2allTicker
        https://doc.xt.com/#futures_market_websocket_v2allAggTicker

        :param str [symbols]: unified market symbols
        :param dict params: extra parameters specific to the xt api endpoint
        :param str [params.method]: 'agg_tickers'(contract only) or 'tickers', default = 'tickers' - the endpoint that will be streamed
        :returns dict: a `ticker structure <https://docs.ccxt.com/en/latest/manual.html#ticker-structure>`
        """
        await self.load_markets()
        options = self.safe_dict(self.options, 'watchTickers')
        defaultMethod = self.safe_string(options, 'method', 'tickers')
        name = self.safe_string(params, 'method', defaultMethod)
        market = None
        if symbols is not None:
            market = self.market(symbols[0])
        tickers = await self.subscribe(name, 'public', 'watchTickers', market, symbols, params)
        if self.newUpdates:
            return tickers
        return self.filter_by_array(self.tickers, 'symbol', symbols)

    async def watch_ohlcv(self, symbol: str, timeframe='1m', since: Int = None, limit: Int = None, params={}) -> List[list]:
        """
        watches historical candlestick data containing the open, high, low, and close price, and the volume of a market

        https://doc.xt.com/#websocket_publicsymbolKline
        https://doc.xt.com/#futures_market_websocket_v2symbolKline

        :param str symbol: unified symbol of the market to fetch OHLCV data for
        :param str timeframe: 1m, 3m, 5m, 15m, 30m, 1h, 2h, 4h, 6h, 8h, 12h, 1d, 3d, 1w, or 1M
        :param int [since]: not used by xt watchOHLCV
        :param int [limit]: not used by xt watchOHLCV
        :param dict params: extra parameters specific to the xt api endpoint
        :returns int[][]: A list of candles ordered, open, high, low, close, volume
        """
        await self.load_markets()
        market = self.market(symbol)
        name = 'kline@' + market['id'] + ',' + timeframe
        ohlcv = await self.subscribe(name, 'public', 'watchOHLCV', market, None, params)
        if self.newUpdates:
            limit = ohlcv.getLimit(symbol, limit)
        return self.filter_by_since_limit(ohlcv, since, limit, 0, True)

    async def watch_trades(self, symbol: str, since: Int = None, limit: Int = None, params={}) -> List[Trade]:
        """
        get the list of most recent trades for a particular symbol

        https://doc.xt.com/#websocket_publicdealRecord
        https://doc.xt.com/#futures_market_websocket_v2dealRecord

        :param str symbol: unified symbol of the market to fetch trades for
        :param int [since]: timestamp in ms of the earliest trade to fetch
        :param int [limit]: the maximum amount of trades to fetch
        :param dict params: extra parameters specific to the xt api endpoint
        :returns dict[]: a list of `trade structures <https://docs.ccxt.com/en/latest/manual.html?#public-trades>`
        """
        await self.load_markets()
        market = self.market(symbol)
        name = 'trade@' + market['id']
        trades = await self.subscribe(name, 'public', 'watchTrades', market, None, params)
        if self.newUpdates:
            limit = trades.getLimit(symbol, limit)
        return self.filter_by_since_limit(trades, since, limit, 'timestamp')

    async def watch_order_book(self, symbol: str, limit: Int = None, params={}) -> OrderBook:
        """
        watches information on open orders with bid(buy) and ask(sell) prices, volumes and other data

        https://doc.xt.com/#websocket_publiclimitDepth
        https://doc.xt.com/#websocket_publicincreDepth
        https://doc.xt.com/#futures_market_websocket_v2limitDepth
        https://doc.xt.com/#futures_market_websocket_v2increDepth

        :param str symbol: unified symbol of the market to fetch the order book for
        :param int [limit]: not used by xt watchOrderBook
        :param dict params: extra parameters specific to the xt api endpoint
        :param int [params.levels]: 5, 10, 20, or 50
        :returns dict: A dictionary of `order book structures <https://docs.ccxt.com/en/latest/manual.html#order-book-structure>` indexed by market symbols
        """
        await self.load_markets()
        market = self.market(symbol)
        levels = self.safe_string(params, 'levels')
        params = self.omit(params, 'levels')
        name = 'depth_update@' + market['id']
        if levels is not None:
            name = 'depth@' + market['id'] + ',' + levels
        orderbook = await self.subscribe(name, 'public', 'watchOrderBook', market, None, params)
        return orderbook.limit()

    async def watch_orders(self, symbol: Str = None, since: Int = None, limit: Int = None, params={}) -> List[Order]:
        """
        watches information on multiple orders made by the user

        https://doc.xt.com/#websocket_privateorderChange
        https://doc.xt.com/#futures_user_websocket_v2order

        :param str [symbol]: unified market symbol
        :param int [since]: not used by xt watchOrders
        :param int [limit]: the maximum number of orders to return
        :param dict params: extra parameters specific to the xt api endpoint
        :returns dict[]: a list of `order structures <https://docs.ccxt.com/en/latest/manual.html#order-structure>`
        """
        await self.load_markets()
        name = 'order'
        market = None
        if symbol is not None:
            market = self.market(symbol)
        orders = await self.subscribe(name, 'private', 'watchOrders', market, None, params)
        if self.newUpdates:
            limit = orders.getLimit(symbol, limit)
        return self.filter_by_since_limit(orders, since, limit, 'timestamp')

    async def watch_my_trades(self, symbol: Str = None, since: Int = None, limit: Int = None, params={}) -> List[Trade]:
        """
        watches information on multiple trades made by the user

        https://doc.xt.com/#websocket_privateorderDeal
        https://doc.xt.com/#futures_user_websocket_v2trade

        :param str symbol: unified market symbol of the market orders were made in
        :param int [since]: the earliest time in ms to fetch orders for
        :param int [limit]: the maximum number of  orde structures to retrieve
        :param dict params: extra parameters specific to the kucoin api endpoint
        :returns dict[]: a list of `trade structures <https://docs.ccxt.com/#/?id=trade-structure>`
        """
        await self.load_markets()
        name = 'trade'
        market = None
        if symbol is not None:
            market = self.market(symbol)
        trades = await self.subscribe(name, 'private', 'watchMyTrades', market, None, params)
        if self.newUpdates:
            limit = trades.getLimit(symbol, limit)
        return self.filter_by_since_limit(trades, since, limit, 'timestamp')

    async def watch_balance(self, params={}) -> Balances:
        """
        watches information on multiple orders made by the user

        https://doc.xt.com/#websocket_privatebalanceChange
        https://doc.xt.com/#futures_user_websocket_v2balance

        :param dict params: extra parameters specific to the xt api endpoint
        :returns dict[]: a list of `balance structures <https://docs.ccxt.com/#/?id=balance-structure>`
        """
        await self.load_markets()
        name = 'balance'
        return await self.subscribe(name, 'private', 'watchBalance', None, None, params)

    async def watch_positions(self, symbols: Strings = None, since: Int = None, limit: Int = None, params={}) -> List[Position]:
        """

        https://doc.xt.com/#futures_user_websocket_v2position

        watch all open positions
        :param str[]|None symbols: list of unified market symbols
        :param number [since]: since timestamp
        :param number [limit]: limit
        :param dict params: extra parameters specific to the exchange API endpoint
        :returns dict[]: a list of `position structure <https://docs.ccxt.com/en/latest/manual.html#position-structure>`
        """
        await self.load_markets()
        url = self.urls['api']['ws']['contract'] + '/' + 'user'
        client = self.client(url)
        self.set_positions_cache(client)
        fetchPositionsSnapshot = self.handle_option('watchPositions', 'fetchPositionsSnapshot', True)
        awaitPositionsSnapshot = self.handle_option('watchPositions', 'awaitPositionsSnapshot', True)
        cache = self.positions
        if fetchPositionsSnapshot and awaitPositionsSnapshot and self.is_empty(cache):
            snapshot = await client.future('fetchPositionsSnapshot')
            return self.filter_by_symbols_since_limit(snapshot, symbols, since, limit, True)
        name = 'position'
        newPositions = await self.subscribe(name, 'private', 'watchPositions', None, None, params)
        if self.newUpdates:
            return newPositions
        return self.filter_by_symbols_since_limit(cache, symbols, since, limit, True)

    def set_positions_cache(self, client: Client):
        if self.positions is None:
            self.positions = ArrayCacheBySymbolBySide()
        fetchPositionsSnapshot = self.handle_option('watchPositions', 'fetchPositionsSnapshot')
        if fetchPositionsSnapshot:
            messageHash = 'fetchPositionsSnapshot'
            if not (messageHash in client.futures):
                client.future(messageHash)
                self.spawn(self.load_positions_snapshot, client, messageHash)

    async def load_positions_snapshot(self, client, messageHash):
        positions = await self.fetch_positions(None)
        self.positions = ArrayCacheBySymbolBySide()
        cache = self.positions
        for i in range(0, len(positions)):
            position = positions[i]
            contracts = self.safe_number(position, 'contracts', 0)
            if contracts > 0:
                cache.append(position)
        # don't remove the future from the .futures cache
        future = client.futures[messageHash]
        future.resolve(cache)
        client.resolve(cache, 'position::contract')

    def handle_position(self, client, message):
        #
        #    {
        #      topic: 'position',
        #      event: 'position',
        #      data: {
        #        accountId: 245296,
        #        accountType: 0,
        #        symbol: 'eth_usdt',
        #        contractType: 'PERPETUAL',
        #        positionType: 'CROSSED',
        #        positionSide: 'LONG',
        #        positionSize: '1',
        #        closeOrderSize: '0',
        #        availableCloseSize: '1',
        #        realizedProfit: '-0.0121',
        #        entryPrice: '2637.87',
        #        openOrderSize: '1',
        #        isolatedMargin: '2.63787',
        #        openOrderMarginFrozen: '2.78832014',
        #        underlyingType: 'U_BASED',
        #        leverage: 10,
        #        welfareAccount: False,
        #        profitFixedLatest: {},
        #        closeProfit: '0.0000',
        #        totalFee: '-0.0158',
        #        totalFundFee: '0.0037',
        #        markPrice: '2690.96'
        #      }
        #    }
        #
        if self.positions is None:
            self.positions = ArrayCacheBySymbolBySide()
        cache = self.positions
        data = self.safe_dict(message, 'data', {})
        position = self.parse_position(data)
        cache.append(position)
        messageHashes = self.find_message_hashes(client, 'position::contract')
        for i in range(0, len(messageHashes)):
            messageHash = messageHashes[i]
            parts = messageHash.split('::')
            symbolsString = parts[1]
            symbols = symbolsString.split(',')
            positions = self.filter_by_array([position], 'symbol', symbols, False)
            if not self.is_empty(positions):
                client.resolve(positions, messageHash)
        client.resolve([position], 'position::contract')

    def handle_ticker(self, client: Client, message: dict):
        #
        # spot
        #
        #    {
        #        topic: 'ticker',
        #        event: 'ticker@btc_usdt',
        #        data: {
        #           s: 'btc_usdt',            # symbol
        #           t: 1683501935877,         # time(Last transaction time)
        #           cv: '-82.67',             # priceChangeValue(24 hour price change)
        #           cr: '-0.0028',            # priceChangeRate 24-hour price change(percentage)
        #           o: '28823.87',            # open price
        #           c: '28741.20',            # close price
        #           h: '29137.64',            # highest price
        #           l: '28660.93',            # lowest price
        #           q: '6372.601573',         # quantity
        #           v: '184086075.2772391'    # volume
        #        }
        #    }
        #
        # contract
        #
        #    {
        #        "topic": "ticker",
        #        "event": "ticker@btc_usdt",
        #        "data": {
        #            "s": "btc_index",  # trading pair
        #            "o": "49000",      # opening price
        #            "c": "50000",      # closing price
        #            "h": "0.1",        # highest price
        #            "l": "0.1",        # lowest price
        #            "a": "0.1",        # volume
        #            "v": "0.1",        # turnover
        #            "ch": "0.21",      # quote change
        #            "t": 123124124     # timestamp
        #       }
        #    }
        #
        # agg_ticker(contract)
        #
        #    {
        #        "topic": "agg_ticker",
        #        "event": "agg_ticker@btc_usdt",
        #        "data": {
        #            "s": "btc_index",          # trading pair
        #            "o": "49000",              # opening price
        #            "c": "50000",              # closing price
        #            "h": "0.1",                # highest price
        #            "l": "0.1",                # lowest price
        #            "a": "0.1",                # volume
        #            "v": "0.1",                # turnover
        #            "ch": "0.21",              # quote change
        #            "i": "0.21" ,              # index price
        #            "m": "0.21",               # mark price
        #            "bp": "0.21",              # bid price
        #            "ap": "0.21" ,             # ask price
        #            "t": 123124124             # timestamp
        #       }
        #    }
        #
        data = self.safe_dict(message, 'data')
        marketId = self.safe_string(data, 's')
        if marketId is not None:
            cv = self.safe_string(data, 'cv')
            isSpot = cv is not None
            ticker = self.parse_ticker(data)
            symbol = ticker['symbol']
            self.tickers[symbol] = ticker
            event = self.safe_string(message, 'event')
            messageHashTail = 'spot' if isSpot else 'contract'
            messageHash = event + '::' + messageHashTail
            client.resolve(ticker, messageHash)
        return message

    def handle_tickers(self, client: Client, message: dict):
        #
        # spot
        #
        #    {
        #        topic: 'tickers',
        #        event: 'tickers',
        #        data: [
        #            {
        #                s: 'elon_usdt',
        #                t: 1683502958381,
        #                cv: '-0.0000000125',
        #                cr: '-0.0495',
        #                o: '0.0000002522',
        #                c: '0.0000002397',
        #                h: '0.0000002690',
        #                l: '0.0000002371',
        #                q: '3803783034.0000000000',
        #                v: '955.3260820022'
        #            },
        #            ...
        #        ]
        #    }
        #
        # contract
        #
        #    {
        #        "topic": "tickers",
        #        "event": "tickers",
        #        "data": [
        #            {
        #                "s": "btc_index",  # trading pair
        #                "o": "49000",      # opening price
        #                "c": "50000",      # closing price
        #                "h": "0.1",        # highest price
        #                "l": "0.1",        # lowest price
        #                "a": "0.1",        # volume
        #                "v": "0.1",        # turnover
        #                "ch": "0.21",      # quote change
        #                "t": 123124124     # timestamp
        #            }
        #        ]
        #    }
        #
        # agg_ticker(contract)
        #
        #    {
        #        "topic": "agg_tickers",
        #        "event": "agg_tickers",
        #        "data": [
        #            {
        #                "s": "btc_index",          # trading pair
        #                "o": "49000",              # opening price
        #                "c": "50000",              # closing price
        #                "h": "0.1",                # highest price
        #                "l": "0.1",                # lowest price
        #                "a": "0.1",                # volume
        #                "v": "0.1",                # turnover
        #                "ch": "0.21",              # quote change
        #                "i": "0.21" ,              # index price
        #                "m": "0.21",               # mark price
        #                "bp": "0.21",              # bid price
        #                "ap": "0.21" ,             # ask price
        #                "t": 123124124             # timestamp
        #            }
        #        ]
        #    }
        #
        data = self.safe_list(message, 'data', [])
        firstTicker = self.safe_dict(data, 0)
        spotTest = self.safe_string_2(firstTicker, 'cv', 'aq')
        tradeType = 'spot' if (spotTest is not None) else 'contract'
        newTickers = []
        for i in range(0, len(data)):
            tickerData = data[i]
            ticker = self.parse_ticker(tickerData)
            symbol = ticker['symbol']
            self.tickers[symbol] = ticker
            newTickers.append(ticker)
        messageHashStart = self.safe_string(message, 'topic') + '::' + tradeType
        messageHashes = self.find_message_hashes(client, messageHashStart + '::')
        for i in range(0, len(messageHashes)):
            messageHash = messageHashes[i]
            parts = messageHash.split('::')
            symbolsString = parts[2]
            symbols = symbolsString.split(',')
            tickers = self.filter_by_array(newTickers, 'symbol', symbols)
            tickersSymbols = list(tickers.keys())
            numTickers = len(tickersSymbols)
            if numTickers > 0:
                client.resolve(tickers, messageHash)
        client.resolve(self.tickers, messageHashStart)
        return message

    def handle_ohlcv(self, client: Client, message: dict):
        #
        # spot
        #
        #    {
        #        "topic": "kline",
        #        "event": "kline@btc_usdt,5m",
        #        "data": {
        #            "s": "btc_usdt",        # symbol
        #            "t": 1656043200000,     # time
        #            "i": "5m",              # interval
        #            "o": "44000",           # open price
        #            "c": "50000",           # close price
        #            "h": "52000",           # highest price
        #            "l": "36000",           # lowest price
        #            "q": "34.2",            # qty(quantity)
        #            "v": "230000"           # volume
        #        }
        #    }
        #
        # contract
        #
        #    {
        #        "topic": "kline",
        #        "event": "kline@btc_usdt,5m",
        #        "data": {
        #            "s": "btc_index",      # trading pair
        #            "o": "49000",          # opening price
        #            "c": "50000",          # closing price
        #            "h": "0.1",            # highest price
        #            "l": "0.1",            # lowest price
        #            "a": "0.1",            # volume
        #            "v": "0.1",            # turnover
        #            "ch": "0.21",          # quote change
        #            "t": 123124124         # timestamp
        #        }
        #    }
        #
        data = self.safe_dict(message, 'data', {})
        marketId = self.safe_string(data, 's')
        if marketId is not None:
            timeframe = self.safe_string(data, 'i')
            tradeType = 'spot' if ('q' in data) else 'contract'
            market = self.safe_market(marketId, None, None, tradeType)
            symbol = market['symbol']
            parsed = self.parse_ohlcv(data, market)
            self.ohlcvs[symbol] = self.safe_dict(self.ohlcvs, symbol, {})
            stored = self.safe_value(self.ohlcvs[symbol], timeframe)
            if stored is None:
                limit = self.safe_integer(self.options, 'OHLCVLimit', 1000)
                stored = ArrayCacheByTimestamp(limit)
                self.ohlcvs[symbol][timeframe] = stored
            stored.append(parsed)
            event = self.safe_string(message, 'event')
            messageHash = event + '::' + tradeType
            client.resolve(stored, messageHash)
        return message

    def handle_trade(self, client: Client, message: dict):
        #
        # spot
        #
        #    {
        #        topic: 'trade',
        #        event: 'trade@btc_usdt',
        #        data: {
        #            s: 'btc_usdt',
        #            i: '228825383103928709',
        #            t: 1684258222702,
        #            p: '27003.65',
        #            q: '0.000796',
        #            b: True
        #        }
        #    }
        #
        # contract
        #
        #    {
        #        "topic": "trade",
        #        "event": "trade@btc_usdt",
        #        "data": {
        #            "s": "btc_index",  # trading pair
        #            "p": "50000",      # price
        #            "a": "0.1"         # Quantity
        #            "m": "BID"         # Deal side  BID:Buy ASK:Sell
        #            "t": 123124124     # timestamp
        #        }
        #    }
        #
        data = self.safe_dict(message, 'data')
        marketId = self.safe_string_lower(data, 's')
        if marketId is not None:
            trade = self.parse_trade(data)
            i = self.safe_string(data, 'i')
            tradeType = 'spot' if (i is not None) else 'contract'
            market = self.safe_market(marketId, None, None, tradeType)
            symbol = market['symbol']
            event = self.safe_string(message, 'event')
            tradesArray = self.safe_value(self.trades, symbol)
            if tradesArray is None:
                tradesLimit = self.safe_integer(self.options, 'tradesLimit', 1000)
                tradesArray = ArrayCache(tradesLimit)
                self.trades[symbol] = tradesArray
            tradesArray.append(trade)
            messageHash = event + '::' + tradeType
            client.resolve(tradesArray, messageHash)
        return message

    def handle_order_book(self, client: Client, message: dict):
        #
        # spot
        #
        #    {
        #        "topic": "depth",
        #        "event": "depth@btc_usdt,20",
        #        "data": {
        #            "s": "btc_usdt",        # symbol
        #            "fi": 1681433733351,    # firstUpdateId = previous lastUpdateId + 1
        #            "i": 1681433733371,     # updateId
        #            "a": [                 # asks(sell order)
        #                [                  # [0]price, [1]quantity
        #                    "34000",        # price
        #                    "1.2"           # quantity
        #                ],
        #                [
        #                    "34001",
        #                    "2.3"
        #                ]
        #            ],
        #            "b": [                  # bids(buy order)
        #                [
        #                    "32000",
        #                    "0.2"
        #                ],
        #                [
        #                    "31000",
        #                    "0.5"
        #                ]
        #            ]
        #        }
        #    }
        #
        # contract
        #
        #    {
        #        "topic": "depth",
        #        "event": "depth@btc_usdt,20",
        #        "data": {
        #            s: "btc_usdt",
        #            pu: "548111455664",
        #            fu: "548111455665",
        #            u: "548111455667",
        #            a: [
        #                [
        #                    "26841.5",
        #                    "50210",
        #                ],
        #            ],
        #            b: [
        #                [
        #                    "26841",
        #                    "67075",
        #                ],
        #            ],
        #            t: 1684530667083,
        #        }
        #    }
        #
        data = self.safe_dict(message, 'data')
        marketId = self.safe_string(data, 's')
        if marketId is not None:
            event = self.safe_string(message, 'event')
            splitEvent = event.split(',')
            event = self.safe_string(splitEvent, 0)
            tradeType = 'contract' if ('fu' in data) else 'spot'
            market = self.safe_market(marketId, None, None, tradeType)
            symbol = market['symbol']
            obAsks = self.safe_list(data, 'a')
            obBids = self.safe_list(data, 'b')
            messageHash = event + '::' + tradeType
            if not (symbol in self.orderbooks):
                subscription = self.safe_dict(client.subscriptions, messageHash, {})
                limit = self.safe_integer(subscription, 'limit')
                self.orderbooks[symbol] = self.order_book({}, limit)
            orderbook = self.orderbooks[symbol]
            nonce = self.safe_integer(orderbook, 'nonce')
            if nonce is None:
                cacheLength = len(orderbook.cache)
                snapshotDelay = self.handle_option('watchOrderBook', 'snapshotDelay', 25)
                if cacheLength == snapshotDelay:
                    self.spawn(self.load_order_book, client, messageHash, symbol)
                orderbook.cache.append(data)
                return
            if obAsks is not None:
                asks = orderbook['asks']
                for i in range(0, len(obAsks)):
                    ask = obAsks[i]
                    price = self.safe_number(ask, 0)
                    quantity = self.safe_number(ask, 1)
                    asks.store(price, quantity)
            if obBids is not None:
                bids = orderbook['bids']
                for i in range(0, len(obBids)):
                    bid = obBids[i]
                    price = self.safe_number(bid, 0)
                    quantity = self.safe_number(bid, 1)
                    bids.store(price, quantity)
            timestamp = self.safe_integer(data, 't')
            orderbook['nonce'] = self.safe_integer_2(data, 'i', 'u')
            orderbook['timestamp'] = timestamp
            orderbook['datetime'] = self.iso8601(timestamp)
            orderbook['symbol'] = symbol
            client.resolve(orderbook, messageHash)

    def parse_ws_order_trade(self, trade: dict, market: Market = None):
        #
        #    {
        #        "s": "btc_usdt",                         # symbol
        #        "t": 1656043204763,                      # time happened time
        #        "i": "6216559590087220004",              # orderId,
        #        "ci": "test123",                         # clientOrderId
        #        "st": "PARTIALLY_FILLED",                # state
        #        "sd": "BUY",                             # side BUY/SELL
        #        "eq": "2",                               # executedQty executed quantity
        #        "ap": "30000",                           # avg price
        #        "f": "0.002"                             # fee
        #    }
        #
        # contract
        #
        #    {
        #        "symbol": "btc_usdt",                    # Trading pair
        #        "orderId": "1234",                       # Order Id
        #        "origQty": "34244",                      # Original Quantity
        #        "avgPrice": "123",                       # Quantity
        #        "price": "1111",                         # Average price
        #        "executedQty": "34244",                  # Volume(Cont)
        #        "orderSide": "BUY",                      # BUY, SELL
        #        "positionSide": "LONG",                  # LONG, SHORT
        #        "marginFrozen": "123",                   # Occupied margin
        #        "sourceType": "default",                 # DEFAULT:normal order,ENTRUST:plan commission,PROFIR:Take Profit and Stop Loss
        #        "sourceId" : "1231231",                  # Triggering conditions ID
        #        "state": "",                             # state:NEW：New order(unfilled);PARTIALLY_FILLED:Partial deal;PARTIALLY_CANCELED:Partial revocation;FILLED:Filled;CANCELED:Cancled;REJECTED:Order failed;EXPIRED：Expired
        #        "createTime": 1731231231,                # CreateTime
        #        "clientOrderId": "204788317630342726"
        #    }
        #
        marketId = self.safe_string(trade, 's')
        tradeType = 'contract' if ('symbol' in trade) else 'spot'
        market = self.safe_market(marketId, market, None, tradeType)
        timestamp = self.safe_string(trade, 't')
        return self.safe_trade({
            'info': trade,
            'id': None,
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'symbol': market['symbol'],
            'order': self.safe_string(trade, 'i', 'orderId'),
            'type': self.parse_order_status(self.safe_string(trade, 'st', 'state')),
            'side': self.safe_string_lower(trade, 'sd', 'orderSide'),
            'takerOrMaker': None,
            'price': self.safe_number(trade, 'price'),
            'amount': self.safe_string(trade, 'origQty'),
            'cost': None,
            'fee': {
                'currency': None,
                'cost': self.safe_number(trade, 'f'),
                'rate': None,
            },
        }, market)

    def parse_ws_order(self, order: dict, market: Market = None):
        #
        # spot
        #
        #    {
        #        "s": "btc_usdt",                # symbol
        #        "bc": "btc",                    # base currency
        #        "qc": "usdt",                   # quotation currency
        #        "t": 1656043204763,             # happened time
        #        "ct": 1656043204663,            # create time
        #        "i": "6216559590087220004",     # order id,
        #        "ci": "test123",                # client order id
        #        "st": "PARTIALLY_FILLED",       # state NEW/PARTIALLY_FILLED/FILLED/CANCELED/REJECTED/EXPIRED
        #        "sd": "BUY",                    # side BUY/SELL
        #        "tp": "LIMIT",                  # type LIMIT/MARKET
        #        "oq":  "4"                      # original quantity
        #        "oqq":  48000,                  # original quotation quantity
        #        "eq": "2",                      # executed quantity
        #        "lq": "2",                      # remaining quantity
        #        "p": "4000",                    # price
        #        "ap": "30000",                  # avg price
        #        "f":"0.002"                     # fee
        #    }
        #
        # contract
        #
        #    {
        #        "symbol": "btc_usdt",                    # Trading pair
        #        "orderId": "1234",                       # Order Id
        #        "origQty": "34244",                      # Original Quantity
        #        "avgPrice": "123",                       # Quantity
        #        "price": "1111",                         # Average price
        #        "executedQty": "34244",                  # Volume(Cont)
        #        "orderSide": "BUY",                      # BUY, SELL
        #        "positionSide": "LONG",                  # LONG, SHORT
        #        "marginFrozen": "123",                   # Occupied margin
        #        "sourceType": "default",                 # DEFAULT:normal order,ENTRUST:plan commission,PROFIR:Take Profit and Stop Loss
        #        "sourceId" : "1231231",                  # Triggering conditions ID
        #        "state": "",                             # state:NEW：New order(unfilled);PARTIALLY_FILLED:Partial deal;PARTIALLY_CANCELED:Partial revocation;FILLED:Filled;CANCELED:Cancled;REJECTED:Order failed;EXPIRED：Expired
        #        "createTime": 1731231231,                # CreateTime
        #        "clientOrderId": "204788317630342726"
        #    }
        #
        marketId = self.safe_string_2(order, 's', 'symbol')
        tradeType = 'contract' if ('symbol' in order) else 'spot'
        market = self.safe_market(marketId, market, None, tradeType)
        timestamp = self.safe_integer_2(order, 'ct', 'createTime')
        return self.safe_order({
            'info': order,
            'id': self.safe_string_2(order, 'i', 'orderId'),
            'clientOrderId': self.safe_string_2(order, 'ci', 'clientOrderId'),
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'lastTradeTimestamp': None,
            'symbol': market['symbol'],
            'type': market['type'],
            'timeInForce': None,
            'postOnly': None,
            'side': self.safe_string_lower_2(order, 'sd', 'orderSide'),
            'price': self.safe_number_2(order, 'p', 'price'),
            'stopPrice': None,
            'stopLoss': None,
            'takeProfit': None,
            'amount': self.safe_string_2(order, 'oq', 'origQty'),
            'filled': self.safe_string_2(order, 'eq', 'executedQty'),
            'remaining': self.safe_string(order, 'lq'),
            'cost': None,
            'average': self.safe_string_2(order, 'ap', 'avgPrice'),
            'status': self.parse_order_status(self.safe_string(order, 'st', 'state')),
            'fee': {
                'currency': None,
                'cost': self.safe_number(order, 'f'),
            },
            'trades': None,
        }, market)

    def handle_order(self, client: Client, message: dict):
        #
        # spot
        #
        #    {
        #        "topic": "order",
        #        "event": "order",
        #        "data": {
        #            "s": "btc_usdt",                # symbol
        #            "t": 1656043204763,             # time happened time
        #            "i": "6216559590087220004",     # orderId,
        #            "ci": "test123",                # clientOrderId
        #            "st": "PARTIALLY_FILLED",       # state
        #            "sd": "BUY",                    # side BUY/SELL
        #            "eq": "2",                      # executedQty executed quantity
        #            "ap": "30000",                  # avg price
        #            "f": "0.002"                    # fee
        #        }
        #    }
        #
        # contract
        #
        #    {
        #        "topic": "order",
        #        "event": "order@123456",
        #        "data": {
        #             "symbol": "btc_usdt",                    # Trading pair
        #             "orderId": "1234",                       # Order Id
        #             "origQty": "34244",                      # Original Quantity
        #             "avgPrice": "123",                       # Quantity
        #             "price": "1111",                         # Average price
        #             "executedQty": "34244",                  # Volume(Cont)
        #             "orderSide": "BUY",                      # BUY, SELL
        #             "positionSide": "LONG",                  # LONG, SHORT
        #             "marginFrozen": "123",                   # Occupied margin
        #             "sourceType": "default",                 # DEFAULT:normal order,ENTRUST:plan commission,PROFIR:Take Profit and Stop Loss
        #             "sourceId" : "1231231",                  # Triggering conditions ID
        #             "state": "",                             # state:NEW：New order(unfilled);PARTIALLY_FILLED:Partial deal;PARTIALLY_CANCELED:Partial revocation;FILLED:Filled;CANCELED:Cancled;REJECTED:Order failed;EXPIRED：Expired
        #             "createTime": 1731231231,                # CreateTime
        #             "clientOrderId": "204788317630342726"
        #           }
        #    }
        #
        orders = self.orders
        if orders is None:
            limit = self.safe_integer(self.options, 'ordersLimit')
            orders = ArrayCacheBySymbolById(limit)
            self.orders = orders
        order = self.safe_dict(message, 'data', {})
        marketId = self.safe_string_2(order, 's', 'symbol')
        if marketId is not None:
            tradeType = 'contract' if ('symbol' in order) else 'spot'
            market = self.safe_market(marketId, None, None, tradeType)
            parsed = self.parse_ws_order(order, market)
            orders.append(parsed)
            client.resolve(orders, 'order::' + tradeType)
        return message

    def handle_balance(self, client: Client, message: dict):
        #
        # spot
        #
        #    {
        #        topic: 'balance',
        #        event: 'balance',
        #        data: {
        #            a: 3513677381884,
        #            t: 1684250056775,
        #            c: 'usdt',
        #            b: '7.71000000',
        #            f: '0.00000000',
        #            z: 'SPOT'
        #        }
        #    }
        #
        # contract
        #
        #    {
        #        "topic": "balance",
        #        "event": "balance@123456",
        #        "data": {
        #            "coin": "usdt",
        #            "underlyingType": 1,                          # 1:Coin-M,2:USDT-M
        #            "walletBalance": "123",                       # Balance
        #            "openOrderMarginFrozen": "123",               # Frozen order
        #            "isolatedMargin": "213",                      # Isolated Margin
        #            "crossedMargin": "0"                          # Crossed Margin
        #            "availableBalance": '2.256114450000000000',
        #            "coupon": '0',
        #            "bonus": '0'
        #        }
        #    }
        #
        data = self.safe_dict(message, 'data', {})
        currencyId = self.safe_string_2(data, 'c', 'coin')
        code = self.safe_currency_code(currencyId)
        account = self.account()
        account['free'] = self.safe_string(data, 'availableBalance')
        account['used'] = self.safe_string(data, 'f')
        account['total'] = self.safe_string_2(data, 'b', 'walletBalance')
        self.balance[code] = account
        self.balance = self.safe_balance(self.balance)
        tradeType = 'contract' if ('coin' in data) else 'spot'
        client.resolve(self.balance, 'balance::' + tradeType)

    def handle_my_trades(self, client: Client, message: dict):
        #
        # spot
        #
        #    {
        #        "topic": "trade",
        #        "event": "trade",
        #        "data": {
        #            "s": "btc_usdt",                # symbol
        #            "t": 1656043204763,             # time
        #            "i": "6316559590087251233",     # tradeId
        #            "oi": "6216559590087220004",    # orderId
        #            "p": "30000",                   # trade price
        #            "q": "3",                       # qty quantity
        #            "v": "90000"                    # volume trade amount
        #        }
        #    }
        #
        # contract
        #
        #    {
        #       "topic": "trade",
        #       "event": "trade@123456",
        #       "data": {
        #            "symbol": 'btc_usdt',
        #            "orderSide": 'SELL',
        #            "positionSide": 'LONG',
        #            "orderId": '231485367663419328',
        #            "price": '27152.7',
        #            "quantity": '33',
        #            "marginUnfrozen": '2.85318000',
        #            "timestamp": 1684892412565
        #        }
        #    }
        #
        data = self.safe_dict(message, 'data', {})
        stored = self.myTrades
        if stored is None:
            limit = self.safe_integer(self.options, 'tradesLimit', 1000)
            stored = ArrayCacheBySymbolById(limit)
            self.myTrades = stored
        parsedTrade = self.parse_trade(data)
        market = self.market(parsedTrade['symbol'])
        stored.append(parsedTrade)
        tradeType = 'contract' if market['contract'] else 'spot'
        client.resolve(stored, 'trade::' + tradeType)

    def handle_message(self, client: Client, message):
        event = self.safe_string(message, 'event')
        if event == 'pong':
            client.onPong()
        elif event is not None:
            topic = self.safe_string(message, 'topic')
            methods = {
                'kline': self.handle_ohlcv,
                'depth': self.handle_order_book,
                'depth_update': self.handle_order_book,
                'ticker': self.handle_ticker,
                'agg_ticker': self.handle_ticker,
                'tickers': self.handle_tickers,
                'agg_tickers': self.handle_tickers,
                'balance': self.handle_balance,
                'order': self.handle_order,
                'position': self.handle_position,
            }
            method = self.safe_value(methods, topic)
            if topic == 'trade':
                data = self.safe_dict(message, 'data')
                if ('oi' in data) or ('orderId' in data):
                    method = self.handle_my_trades
                else:
                    method = self.handle_trade
            if method is not None:
                method(client, message)

    def ping(self, client: Client):
        client.lastPong = self.milliseconds()
        return 'ping'

    def handle_error_message(self, client: Client, message: dict):
        #
        #    {
        #        "id": "123",
        #        "code": 401,
        #        "msg": "token expire"
        #    }
        #
        msg = self.safe_string(message, 'msg')
        if (msg == 'invalid_listen_key') or (msg == 'token expire'):
            client.subscriptions['token'] = None
            self.get_listen_key(True)
            return
        client.reject(message)



from ccxt.base.types import Entry


class ImplicitAPI:
    public_get_get_all_currencies = publicGetGetAllCurrencies = Entry('get_all_currencies', 'public', 'GET', {})
    public_post_build_register_session_key_tx = publicPostBuildRegisterSessionKeyTx = Entry('build_register_session_key_tx', 'public', 'POST', {})
    public_post_register_session_key = publicPostRegisterSessionKey = Entry('register_session_key', 'public', 'POST', {})
    public_post_deregister_session_key = publicPostDeregisterSessionKey = Entry('deregister_session_key', 'public', 'POST', {})
    public_post_login = publicPostLogin = Entry('login', 'public', 'POST', {})
    public_post_statistics = publicPostStatistics = Entry('statistics', 'public', 'POST', {})
    public_post_get_all_currencies = publicPostGetAllCurrencies = Entry('get_all_currencies', 'public', 'POST', {})
    public_post_get_currency = publicPostGetCurrency = Entry('get_currency', 'public', 'POST', {})
    public_post_get_instrument = publicPostGetInstrument = Entry('get_instrument', 'public', 'POST', {})
    public_post_get_all_instruments = publicPostGetAllInstruments = Entry('get_all_instruments', 'public', 'POST', {})
    public_post_get_instruments = publicPostGetInstruments = Entry('get_instruments', 'public', 'POST', {})
    public_post_get_ticker = publicPostGetTicker = Entry('get_ticker', 'public', 'POST', {})
    public_post_get_latest_signed_feeds = publicPostGetLatestSignedFeeds = Entry('get_latest_signed_feeds', 'public', 'POST', {})
    public_post_get_option_settlement_prices = publicPostGetOptionSettlementPrices = Entry('get_option_settlement_prices', 'public', 'POST', {})
    public_post_get_spot_feed_history = publicPostGetSpotFeedHistory = Entry('get_spot_feed_history', 'public', 'POST', {})
    public_post_get_spot_feed_history_candles = publicPostGetSpotFeedHistoryCandles = Entry('get_spot_feed_history_candles', 'public', 'POST', {})
    public_post_get_funding_rate_history = publicPostGetFundingRateHistory = Entry('get_funding_rate_history', 'public', 'POST', {})
    public_post_get_trade_history = publicPostGetTradeHistory = Entry('get_trade_history', 'public', 'POST', {})
    public_post_get_option_settlement_history = publicPostGetOptionSettlementHistory = Entry('get_option_settlement_history', 'public', 'POST', {})
    public_post_get_liquidation_history = publicPostGetLiquidationHistory = Entry('get_liquidation_history', 'public', 'POST', {})
    public_post_get_interest_rate_history = publicPostGetInterestRateHistory = Entry('get_interest_rate_history', 'public', 'POST', {})
    public_post_get_transaction = publicPostGetTransaction = Entry('get_transaction', 'public', 'POST', {})
    public_post_get_margin = publicPostGetMargin = Entry('get_margin', 'public', 'POST', {})
    public_post_margin_watch = publicPostMarginWatch = Entry('margin_watch', 'public', 'POST', {})
    public_post_validate_invite_code = publicPostValidateInviteCode = Entry('validate_invite_code', 'public', 'POST', {})
    public_post_get_points = publicPostGetPoints = Entry('get_points', 'public', 'POST', {})
    public_post_get_all_points = publicPostGetAllPoints = Entry('get_all_points', 'public', 'POST', {})
    public_post_get_points_leaderboard = publicPostGetPointsLeaderboard = Entry('get_points_leaderboard', 'public', 'POST', {})
    public_post_get_descendant_tree = publicPostGetDescendantTree = Entry('get_descendant_tree', 'public', 'POST', {})
    public_post_get_tree_roots = publicPostGetTreeRoots = Entry('get_tree_roots', 'public', 'POST', {})
    public_post_get_swell_percent_points = publicPostGetSwellPercentPoints = Entry('get_swell_percent_points', 'public', 'POST', {})
    public_post_get_vault_assets = publicPostGetVaultAssets = Entry('get_vault_assets', 'public', 'POST', {})
    public_post_get_etherfi_effective_balances = publicPostGetEtherfiEffectiveBalances = Entry('get_etherfi_effective_balances', 'public', 'POST', {})
    public_post_get_kelp_effective_balances = publicPostGetKelpEffectiveBalances = Entry('get_kelp_effective_balances', 'public', 'POST', {})
    public_post_get_bridge_balances = publicPostGetBridgeBalances = Entry('get_bridge_balances', 'public', 'POST', {})
    public_post_get_ethena_participants = publicPostGetEthenaParticipants = Entry('get_ethena_participants', 'public', 'POST', {})
    public_post_get_vault_share = publicPostGetVaultShare = Entry('get_vault_share', 'public', 'POST', {})
    public_post_get_vault_statistics = publicPostGetVaultStatistics = Entry('get_vault_statistics', 'public', 'POST', {})
    public_post_get_vault_balances = publicPostGetVaultBalances = Entry('get_vault_balances', 'public', 'POST', {})
    public_post_estimate_integrator_points = publicPostEstimateIntegratorPoints = Entry('estimate_integrator_points', 'public', 'POST', {})
    public_post_create_subaccount_debug = publicPostCreateSubaccountDebug = Entry('create_subaccount_debug', 'public', 'POST', {})
    public_post_deposit_debug = publicPostDepositDebug = Entry('deposit_debug', 'public', 'POST', {})
    public_post_withdraw_debug = publicPostWithdrawDebug = Entry('withdraw_debug', 'public', 'POST', {})
    public_post_send_quote_debug = publicPostSendQuoteDebug = Entry('send_quote_debug', 'public', 'POST', {})
    public_post_execute_quote_debug = publicPostExecuteQuoteDebug = Entry('execute_quote_debug', 'public', 'POST', {})
    public_post_get_invite_code = publicPostGetInviteCode = Entry('get_invite_code', 'public', 'POST', {})
    public_post_register_invite = publicPostRegisterInvite = Entry('register_invite', 'public', 'POST', {})
    public_post_get_time = publicPostGetTime = Entry('get_time', 'public', 'POST', {})
    public_post_get_live_incidents = publicPostGetLiveIncidents = Entry('get_live_incidents', 'public', 'POST', {})
    public_post_get_maker_programs = publicPostGetMakerPrograms = Entry('get_maker_programs', 'public', 'POST', {})
    public_post_get_maker_program_scores = publicPostGetMakerProgramScores = Entry('get_maker_program_scores', 'public', 'POST', {})
    private_post_get_account = privatePostGetAccount = Entry('get_account', 'private', 'POST', {})
    private_post_create_subaccount = privatePostCreateSubaccount = Entry('create_subaccount', 'private', 'POST', {})
    private_post_get_subaccount = privatePostGetSubaccount = Entry('get_subaccount', 'private', 'POST', {})
    private_post_get_subaccounts = privatePostGetSubaccounts = Entry('get_subaccounts', 'private', 'POST', {})
    private_post_get_all_portfolios = privatePostGetAllPortfolios = Entry('get_all_portfolios', 'private', 'POST', {})
    private_post_change_subaccount_label = privatePostChangeSubaccountLabel = Entry('change_subaccount_label', 'private', 'POST', {})
    private_post_get_notificationsv = privatePostGetNotificationsv = Entry('get_notificationsv', 'private', 'POST', {})
    private_post_update_notifications = privatePostUpdateNotifications = Entry('update_notifications', 'private', 'POST', {})
    private_post_deposit = privatePostDeposit = Entry('deposit', 'private', 'POST', {})
    private_post_withdraw = privatePostWithdraw = Entry('withdraw', 'private', 'POST', {})
    private_post_transfer_erc20 = privatePostTransferErc20 = Entry('transfer_erc20', 'private', 'POST', {})
    private_post_transfer_position = privatePostTransferPosition = Entry('transfer_position', 'private', 'POST', {})
    private_post_transfer_positions = privatePostTransferPositions = Entry('transfer_positions', 'private', 'POST', {})
    private_post_order = privatePostOrder = Entry('order', 'private', 'POST', {})
    private_post_replace = privatePostReplace = Entry('replace', 'private', 'POST', {})
    private_post_order_debug = privatePostOrderDebug = Entry('order_debug', 'private', 'POST', {})
    private_post_get_order = privatePostGetOrder = Entry('get_order', 'private', 'POST', {})
    private_post_get_orders = privatePostGetOrders = Entry('get_orders', 'private', 'POST', {})
    private_post_get_open_orders = privatePostGetOpenOrders = Entry('get_open_orders', 'private', 'POST', {})
    private_post_cancel = privatePostCancel = Entry('cancel', 'private', 'POST', {})
    private_post_cancel_by_label = privatePostCancelByLabel = Entry('cancel_by_label', 'private', 'POST', {})
    private_post_cancel_by_nonce = privatePostCancelByNonce = Entry('cancel_by_nonce', 'private', 'POST', {})
    private_post_cancel_by_instrument = privatePostCancelByInstrument = Entry('cancel_by_instrument', 'private', 'POST', {})
    private_post_cancel_all = privatePostCancelAll = Entry('cancel_all', 'private', 'POST', {})
    private_post_cancel_trigger_order = privatePostCancelTriggerOrder = Entry('cancel_trigger_order', 'private', 'POST', {})
    private_post_get_order_history = privatePostGetOrderHistory = Entry('get_order_history', 'private', 'POST', {})
    private_post_get_trade_history = privatePostGetTradeHistory = Entry('get_trade_history', 'private', 'POST', {})
    private_post_get_deposit_history = privatePostGetDepositHistory = Entry('get_deposit_history', 'private', 'POST', {})
    private_post_get_withdrawal_history = privatePostGetWithdrawalHistory = Entry('get_withdrawal_history', 'private', 'POST', {})
    private_post_send_rfq = privatePostSendRfq = Entry('send_rfq', 'private', 'POST', {})
    private_post_cancel_rfq = privatePostCancelRfq = Entry('cancel_rfq', 'private', 'POST', {})
    private_post_cancel_batch_rfqs = privatePostCancelBatchRfqs = Entry('cancel_batch_rfqs', 'private', 'POST', {})
    private_post_get_rfqs = privatePostGetRfqs = Entry('get_rfqs', 'private', 'POST', {})
    private_post_poll_rfqs = privatePostPollRfqs = Entry('poll_rfqs', 'private', 'POST', {})
    private_post_send_quote = privatePostSendQuote = Entry('send_quote', 'private', 'POST', {})
    private_post_cancel_quote = privatePostCancelQuote = Entry('cancel_quote', 'private', 'POST', {})
    private_post_cancel_batch_quotes = privatePostCancelBatchQuotes = Entry('cancel_batch_quotes', 'private', 'POST', {})
    private_post_get_quotes = privatePostGetQuotes = Entry('get_quotes', 'private', 'POST', {})
    private_post_poll_quotes = privatePostPollQuotes = Entry('poll_quotes', 'private', 'POST', {})
    private_post_execute_quote = privatePostExecuteQuote = Entry('execute_quote', 'private', 'POST', {})
    private_post_rfq_get_best_quote = privatePostRfqGetBestQuote = Entry('rfq_get_best_quote', 'private', 'POST', {})
    private_post_get_margin = privatePostGetMargin = Entry('get_margin', 'private', 'POST', {})
    private_post_get_collaterals = privatePostGetCollaterals = Entry('get_collaterals', 'private', 'POST', {})
    private_post_get_positions = privatePostGetPositions = Entry('get_positions', 'private', 'POST', {})
    private_post_get_option_settlement_history = privatePostGetOptionSettlementHistory = Entry('get_option_settlement_history', 'private', 'POST', {})
    private_post_get_subaccount_value_history = privatePostGetSubaccountValueHistory = Entry('get_subaccount_value_history', 'private', 'POST', {})
    private_post_expired_and_cancelled_history = privatePostExpiredAndCancelledHistory = Entry('expired_and_cancelled_history', 'private', 'POST', {})
    private_post_get_funding_history = privatePostGetFundingHistory = Entry('get_funding_history', 'private', 'POST', {})
    private_post_get_interest_history = privatePostGetInterestHistory = Entry('get_interest_history', 'private', 'POST', {})
    private_post_get_erc20_transfer_history = privatePostGetErc20TransferHistory = Entry('get_erc20_transfer_history', 'private', 'POST', {})
    private_post_get_liquidation_history = privatePostGetLiquidationHistory = Entry('get_liquidation_history', 'private', 'POST', {})
    private_post_liquidate = privatePostLiquidate = Entry('liquidate', 'private', 'POST', {})
    private_post_get_liquidator_history = privatePostGetLiquidatorHistory = Entry('get_liquidator_history', 'private', 'POST', {})
    private_post_session_keys = privatePostSessionKeys = Entry('session_keys', 'private', 'POST', {})
    private_post_edit_session_key = privatePostEditSessionKey = Entry('edit_session_key', 'private', 'POST', {})
    private_post_register_scoped_session_key = privatePostRegisterScopedSessionKey = Entry('register_scoped_session_key', 'private', 'POST', {})
    private_post_get_mmp_config = privatePostGetMmpConfig = Entry('get_mmp_config', 'private', 'POST', {})
    private_post_set_mmp_config = privatePostSetMmpConfig = Entry('set_mmp_config', 'private', 'POST', {})
    private_post_reset_mmp = privatePostResetMmp = Entry('reset_mmp', 'private', 'POST', {})
    private_post_set_cancel_on_disconnect = privatePostSetCancelOnDisconnect = Entry('set_cancel_on_disconnect', 'private', 'POST', {})
    private_post_get_invite_code = privatePostGetInviteCode = Entry('get_invite_code', 'private', 'POST', {})
    private_post_register_invite = privatePostRegisterInvite = Entry('register_invite', 'private', 'POST', {})



# -*- coding: utf-8 -*-

# PLEASE DO NOT EDIT THIS FILE, IT IS GENERATED AND WILL BE OVERWRITTEN:
# https://github.com/ccxt/ccxt/blob/master/CONTRIBUTING.md#how-to-contribute-code

import ccxt.async_support
from ccxt.async_support.base.ws.cache import ArrayCache, ArrayCacheBySymbolById, ArrayCacheBySymbolBySide, ArrayCacheByTimestamp
import hashlib
from ccxt.base.types import Any, Balances, Int, Market, Order, OrderBook, Position, Str, Strings, Ticker, Tickers, Trade
from ccxt.async_support.base.ws.client import Client
from typing import List
from ccxt.base.errors import ExchangeError
from ccxt.base.errors import ArgumentsRequired
from ccxt.base.errors import NotSupported


class blofin(ccxt.async_support.blofin):

    def describe(self) -> Any:
        return self.deep_extend(super(blofin, self).describe(), {
            'has': {
                'ws': True,
                'watchTrades': True,
                'watchTradesForSymbols': True,
                'watchOrderBook': True,
                'watchOrderBookForSymbols': True,
                'watchTicker': True,
                'watchTickers': True,
                'watchBidsAsks': True,
                'watchOHLCV': True,
                'watchOHLCVForSymbols': True,
                'watchOrders': True,
                'watchOrdersForSymbols': True,
                'watchPositions': True,
            },
            'urls': {
                'api': {
                    'ws': {
                        'swap': {
                            'public': 'wss://openapi.blofin.com/ws/public',
                            'private': 'wss://openapi.blofin.com/ws/private',
                        },
                    },
                },
                'test': {
                    'ws': {
                        'swap': {
                            'public': 'wss://demo-trading-openapi.blofin.com/ws/public',
                            'private': 'wss://demo-trading-openapi.blofin.com/ws/private',
                        },
                    },
                },
            },
            'options': {
                'defaultType': 'swap',
                'tradesLimit': 1000,
                # orderbook channel can be one from:
                #  - "books": 200 depth levels will be pushed in the initial full snapshot. Incremental data will be pushed every 100 ms for the changes in the order book during that period of time.
                #  - "books5": 5 depth levels snapshot will be pushed every time. Snapshot data will be pushed every 100 ms when there are changes in the 5 depth levels snapshot.
                'watchOrderBook': {
                    'channel': 'books',
                },
                'watchOrderBookForSymbols': {
                    'channel': 'books',
                },
            },
            'streaming': {
                'ping': self.ping,
                'keepAlive': 25000,  # 30 seconds max
            },
        })

    def ping(self, client):
        return 'ping'

    def handle_pong(self, client: Client, message):
        #
        #   'pong'
        #
        client.lastPong = self.milliseconds()

    async def watch_trades(self, symbol: str, since: Int = None, limit: Int = None, params={}) -> List[Trade]:
        """
        get the list of most recent trades for a particular symbol

        https://docs.blofin.com/index.html#ws-trades-channel

        :param str symbol: unified symbol of the market to fetch trades for
        :param int [since]: timestamp in ms of the earliest trade to fetch
        :param int [limit]: the maximum amount of trades to fetch
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict[]: a list of `trade structures <https://docs.ccxt.com/#/?id=public-trades>`
        """
        params['callerMethodName'] = 'watchTrades'
        return await self.watch_trades_for_symbols([symbol], since, limit, params)

    async def watch_trades_for_symbols(self, symbols: List[str], since: Int = None, limit: Int = None, params={}) -> List[Trade]:
        """
        get the list of most recent trades for a list of symbols

        https://docs.blofin.com/index.html#ws-trades-channel

        :param str[] symbols: unified symbol of the market to fetch trades for
        :param int [since]: timestamp in ms of the earliest trade to fetch
        :param int [limit]: the maximum amount of trades to fetch
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict[]: a list of `trade structures <https://docs.ccxt.com/#/?id=public-trades>`
        """
        await self.load_markets()
        trades = await self.watch_multiple_wrapper(True, 'trades', 'watchTradesForSymbols', symbols, params)
        if self.newUpdates:
            firstMarket = self.safe_dict(trades, 0)
            firstSymbol = self.safe_string(firstMarket, 'symbol')
            limit = trades.getLimit(firstSymbol, limit)
        return self.filter_by_since_limit(trades, since, limit, 'timestamp', True)

    def handle_trades(self, client: Client, message):
        #
        #     {
        #       arg: {
        #         channel: "trades",
        #         instId: "DOGE-USDT",
        #       },
        #       data : [
        #         <same object in REST example>,
        #         ...
        #       ]
        #     }
        #
        arg = self.safe_dict(message, 'arg')
        channelName = self.safe_string(arg, 'channel')
        data = self.safe_list(message, 'data')
        if data is None:
            return
        for i in range(0, len(data)):
            rawTrade = data[i]
            trade = self.parse_ws_trade(rawTrade)
            symbol = trade['symbol']
            stored = self.safe_value(self.trades, symbol)
            if stored is None:
                limit = self.safe_integer(self.options, 'tradesLimit', 1000)
                stored = ArrayCache(limit)
                self.trades[symbol] = stored
            stored.append(trade)
            messageHash = channelName + ':' + symbol
            client.resolve(stored, messageHash)

    def parse_ws_trade(self, trade, market: Market = None) -> Trade:
        return self.parse_trade(trade, market)

    async def watch_order_book(self, symbol: str, limit: Int = None, params={}) -> OrderBook:
        """
        watches information on open orders with bid(buy) and ask(sell) prices, volumes and other data

        https://docs.blofin.com/index.html#ws-order-book-channel

        :param str symbol: unified symbol of the market to fetch the order book for
        :param int [limit]: the maximum amount of order book entries to return
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: A dictionary of `order book structures <https://docs.ccxt.com/#/?id=order-book-structure>` indexed by market symbols
        """
        params['callerMethodName'] = 'watchOrderBook'
        return await self.watch_order_book_for_symbols([symbol], limit, params)

    async def watch_order_book_for_symbols(self, symbols: List[str], limit: Int = None, params={}) -> OrderBook:
        """
        watches information on open orders with bid(buy) and ask(sell) prices, volumes and other data

        https://docs.blofin.com/index.html#ws-order-book-channel

        :param str[] symbols: unified array of symbols
        :param int [limit]: the maximum amount of order book entries to return
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :param str [params.depth]: the type of order book to subscribe to, default is 'depth/increase100', also accepts 'depth5' or 'depth20' or depth50
        :returns dict: A dictionary of `order book structures <https://docs.ccxt.com/#/?id=order-book-structure>` indexed by market symbols
        """
        await self.load_markets()
        callerMethodName = None
        callerMethodName, params = self.handle_param_string(params, 'callerMethodName', 'watchOrderBookForSymbols')
        channelName = None
        channelName, params = self.handle_option_and_params(params, callerMethodName, 'channel', 'books')
        # due to some problem, temporarily disable other channels
        if channelName != 'books':
            raise NotSupported(self.id + ' ' + callerMethodName + '() at self moment ' + channelName + ' is not supported, coming soon')
        orderbook = await self.watch_multiple_wrapper(True, channelName, callerMethodName, symbols, params)
        return orderbook.limit()

    def handle_order_book(self, client: Client, message):
        #
        #   {
        #     arg: {
        #         channel: "books",
        #         instId: "DOGE-USDT",
        #     },
        #     action: "snapshot",  # can be 'snapshot' or 'update'
        #     data: {
        #         asks: [  [0.08096, 1], [0.08097, 123], ...   ],
        #         bids: [  [0.08095, 4], [0.08094, 237], ...   ],
        #         ts: "1707491587909",
        #         prevSeqId: "0",  # in case of 'update' there will be some value, less then seqId
        #         seqId: "3374250786",
        #     },
        # }
        #
        arg = self.safe_dict(message, 'arg')
        channelName = self.safe_string(arg, 'channel')
        data = self.safe_dict(message, 'data')
        marketId = self.safe_string(arg, 'instId')
        market = self.safe_market(marketId)
        symbol = market['symbol']
        messageHash = channelName + ':' + symbol
        if not (symbol in self.orderbooks):
            self.orderbooks[symbol] = self.order_book()
        orderbook = self.orderbooks[symbol]
        timestamp = self.safe_integer(data, 'ts')
        action = self.safe_string(message, 'action')
        if action == 'snapshot':
            orderBookSnapshot = self.parse_order_book(data, symbol, timestamp)
            orderBookSnapshot['nonce'] = self.safe_integer(data, 'seqId')
            orderbook.reset(orderBookSnapshot)
        else:
            asks = self.safe_list(data, 'asks', [])
            bids = self.safe_list(data, 'bids', [])
            self.handle_deltas_with_keys(orderbook['asks'], asks)
            self.handle_deltas_with_keys(orderbook['bids'], bids)
            orderbook['timestamp'] = timestamp
            orderbook['datetime'] = self.iso8601(timestamp)
        self.orderbooks[symbol] = orderbook
        client.resolve(orderbook, messageHash)

    async def watch_ticker(self, symbol: str, params={}) -> Ticker:
        """
        watches a price ticker, a statistical calculation with the information calculated over the past 24 hours for a specific market

        https://docs.blofin.com/index.html#ws-tickers-channel

        :param str symbol: unified symbol of the market to fetch the ticker for
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: a `ticker structure <https://docs.ccxt.com/#/?id=ticker-structure>`
        """
        params['callerMethodName'] = 'watchTicker'
        market = self.market(symbol)
        symbol = market['symbol']
        result = await self.watch_tickers([symbol], params)
        return result[symbol]

    async def watch_tickers(self, symbols: Strings = None, params={}) -> Tickers:
        """
        watches a price ticker, a statistical calculation with the information calculated over the past 24 hours for all markets of a specific list

        https://docs.blofin.com/index.html#ws-tickers-channel

        :param str[] symbols: unified symbol of the market to fetch the ticker for
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: a `ticker structure <https://docs.ccxt.com/#/?id=ticker-structure>`
        """
        if symbols is None:
            raise NotSupported(self.id + ' watchTickers() requires a list of symbols')
        ticker = await self.watch_multiple_wrapper(True, 'tickers', 'watchTickers', symbols, params)
        if self.newUpdates:
            tickers = {}
            tickers[ticker['symbol']] = ticker
            return tickers
        return self.filter_by_array(self.tickers, 'symbol', symbols)

    def handle_ticker(self, client: Client, message):
        #
        # message
        #
        #     {
        #         arg: {
        #             channel: "tickers",
        #             instId: "DOGE-USDT",
        #         },
        #         data: [
        #             <same object in REST example>
        #         ],
        #     }
        #
        self.handle_bid_ask(client, message)
        arg = self.safe_dict(message, 'arg')
        channelName = self.safe_string(arg, 'channel')
        data = self.safe_list(message, 'data')
        for i in range(0, len(data)):
            ticker = self.parse_ws_ticker(data[i])
            symbol = ticker['symbol']
            messageHash = channelName + ':' + symbol
            self.tickers[symbol] = ticker
            client.resolve(self.tickers[symbol], messageHash)

    def parse_ws_ticker(self, ticker, market: Market = None) -> Ticker:
        return self.parse_ticker(ticker, market)

    async def watch_bids_asks(self, symbols: Strings = None, params={}) -> Tickers:
        """
        watches best bid & ask for symbols

        https://docs.blofin.com/index.html#ws-tickers-channel

        :param str[] symbols: unified symbol of the market to fetch the ticker for
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: a `ticker structure <https://docs.ccxt.com/#/?id=ticker-structure>`
        """
        await self.load_markets()
        symbols = self.market_symbols(symbols, None, False)
        firstMarket = self.market(symbols[0])
        channel = 'tickers'
        marketType = None
        marketType, params = self.handle_market_type_and_params('watchBidsAsks', firstMarket, params)
        url = self.implode_hostname(self.urls['api']['ws'][marketType]['public'])
        messageHashes = []
        args = []
        for i in range(0, len(symbols)):
            market = self.market(symbols[i])
            messageHashes.append('bidask:' + market['symbol'])
            args.append({
                'channel': channel,
                'instId': market['id'],
            })
        request = self.get_subscription_request(args)
        ticker = await self.watch_multiple(url, messageHashes, self.deep_extend(request, params), messageHashes)
        if self.newUpdates:
            tickers = {}
            tickers[ticker['symbol']] = ticker
            return tickers
        return self.filter_by_array(self.bidsasks, 'symbol', symbols)

    def handle_bid_ask(self, client: Client, message):
        data = self.safe_list(message, 'data')
        for i in range(0, len(data)):
            ticker = self.parse_ws_bid_ask(data[i])
            symbol = ticker['symbol']
            messageHash = 'bidask:' + symbol
            self.bidsasks[symbol] = ticker
            client.resolve(ticker, messageHash)

    def parse_ws_bid_ask(self, ticker, market=None):
        marketId = self.safe_string(ticker, 'instId')
        market = self.safe_market(marketId, market, '-')
        symbol = self.safe_string(market, 'symbol')
        timestamp = self.safe_integer(ticker, 'ts')
        return self.safe_ticker({
            'symbol': symbol,
            'timestamp': timestamp,
            'datetime': self.iso8601(timestamp),
            'ask': self.safe_string(ticker, 'askPrice'),
            'askVolume': self.safe_string(ticker, 'askSize'),
            'bid': self.safe_string(ticker, 'bidPrice'),
            'bidVolume': self.safe_string(ticker, 'bidSize'),
            'info': ticker,
        }, market)

    async def watch_ohlcv(self, symbol: str, timeframe='1m', since: Int = None, limit: Int = None, params={}) -> List[list]:
        """
        watches historical candlestick data containing the open, high, low, and close price, and the volume of a market
        :param str symbol: unified symbol of the market to fetch OHLCV data for
        :param str timeframe: the length of time each candle represents
        :param int [since]: timestamp in ms of the earliest candle to fetch
        :param int [limit]: the maximum amount of candles to fetch
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns int[][]: A list of candles ordered, open, high, low, close, volume
        """
        params['callerMethodName'] = 'watchOHLCV'
        result = await self.watch_ohlcv_for_symbols([[symbol, timeframe]], since, limit, params)
        return result[symbol][timeframe]

    async def watch_ohlcv_for_symbols(self, symbolsAndTimeframes: List[List[str]], since: Int = None, limit: Int = None, params={}):
        """
        watches historical candlestick data containing the open, high, low, and close price, and the volume of a market

        https://docs.blofin.com/index.html#ws-candlesticks-channel

        :param str[][] symbolsAndTimeframes: array of arrays containing unified symbols and timeframes to fetch OHLCV data for, example [['BTC/USDT', '1m'], ['LTC/USDT', '5m']]
        :param int [since]: timestamp in ms of the earliest candle to fetch
        :param int [limit]: the maximum amount of candles to fetch
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns int[][]: A list of candles ordered, open, high, low, close, volume
        """
        symbolsLength = len(symbolsAndTimeframes)
        if symbolsLength == 0 or not isinstance(symbolsAndTimeframes[0], list):
            raise ArgumentsRequired(self.id + " watchOHLCVForSymbols() requires a an array of symbols and timeframes, like  [['BTC/USDT', '1m'], ['LTC/USDT', '5m']]")
        await self.load_markets()
        symbol, timeframe, candles = await self.watch_multiple_wrapper(True, 'candle', 'watchOHLCVForSymbols', symbolsAndTimeframes, params)
        if self.newUpdates:
            limit = candles.getLimit(symbol, limit)
        filtered = self.filter_by_since_limit(candles, since, limit, 0, True)
        return self.create_ohlcv_object(symbol, timeframe, filtered)

    def handle_ohlcv(self, client: Client, message):
        #
        # message
        #
        #     {
        #         arg: {
        #             channel: "candle1m",
        #             instId: "DOGE-USDT",
        #         },
        #         data: [
        #             [same object in REST example]
        #         ],
        #     }
        #
        arg = self.safe_dict(message, 'arg')
        channelName = self.safe_string(arg, 'channel')
        data = self.safe_list(message, 'data')
        marketId = self.safe_string(arg, 'instId')
        market = self.safe_market(marketId)
        symbol = market['symbol']
        interval = channelName.replace('candle', '')
        unifiedTimeframe = self.find_timeframe(interval)
        self.ohlcvs[symbol] = self.safe_dict(self.ohlcvs, symbol, {})
        stored = self.safe_value(self.ohlcvs[symbol], unifiedTimeframe)
        if stored is None:
            limit = self.safe_integer(self.options, 'OHLCVLimit', 1000)
            stored = ArrayCacheByTimestamp(limit)
            self.ohlcvs[symbol][unifiedTimeframe] = stored
        for i in range(0, len(data)):
            candle = data[i]
            parsed = self.parse_ohlcv(candle, market)
            stored.append(parsed)
        resolveData = [symbol, unifiedTimeframe, stored]
        messageHash = 'candle' + interval + ':' + symbol
        client.resolve(resolveData, messageHash)

    async def watch_balance(self, params={}) -> Balances:
        """
        query for balance and get the amount of funds available for trading or funds locked in orders

        https://docs.blofin.com/index.html#ws-account-channel

        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict: a `balance structure <https://docs.ccxt.com/#/?id=balance-structure>`
        """
        await self.load_markets()
        await self.authenticate()
        marketType = None
        marketType, params = self.handle_market_type_and_params('watchBalance', None, params)
        if marketType == 'spot':
            raise NotSupported(self.id + ' watchBalance() is not supported for spot markets yet')
        messageHash = marketType + ':balance'
        sub = {
            'channel': 'account',
        }
        request = self.get_subscription_request([sub])
        url = self.implode_hostname(self.urls['api']['ws'][marketType]['private'])
        return await self.watch(url, messageHash, self.deep_extend(request, params), messageHash)

    def handle_balance(self, client: Client, message):
        #
        #     {
        #         arg: {
        #           channel: "account",
        #         },
        #         data: <same object in REST example>,
        #     }
        #
        marketType = 'swap'  # for now
        if not (marketType in self.balance):
            self.balance[marketType] = {}
        self.balance[marketType] = self.parse_ws_balance(message)
        messageHash = marketType + ':balance'
        client.resolve(self.balance[marketType], messageHash)

    def parse_ws_balance(self, message):
        return self.parse_balance(message)

    async def watch_orders(self, symbol: Str = None, since: Int = None, limit: Int = None, params={}) -> List[Order]:
        """
        watches information on multiple orders made by the user
        :param str symbol: unified market symbol of the market orders were made in
        :param int [since]: the earliest time in ms to fetch orders for
        :param int [limit]: the maximum number of order structures to retrieve
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict[]: a list of [order structures]{@link https://docs.ccxt.com/#/?id=order-structure
        """
        params['callerMethodName'] = 'watchOrders'
        symbolsArray = [symbol] if (symbol is not None) else []
        return await self.watch_orders_for_symbols(symbolsArray, since, limit, params)

    async def watch_orders_for_symbols(self, symbols: List[str], since: Int = None, limit: Int = None, params={}) -> List[Order]:
        """
        watches information on multiple orders made by the user across multiple symbols

        https://docs.blofin.com/index.html#ws-order-channel

        :param str[] symbols:
        :param int [since]: the earliest time in ms to fetch orders for
        :param int [limit]: the maximum number of order structures to retrieve
        :param dict [params]: extra parameters specific to the exchange API endpoint
        :returns dict[]: a list of [order structures]{@link https://docs.ccxt.com/#/?id=order-structure
        """
        await self.authenticate()
        await self.load_markets()
        orders = await self.watch_multiple_wrapper(False, 'orders', 'watchOrdersForSymbols', symbols, params)
        if self.newUpdates:
            first = self.safe_value(orders, 0)
            tradeSymbol = self.safe_string(first, 'symbol')
            limit = orders.getLimit(tradeSymbol, limit)
        return self.filter_by_since_limit(orders, since, limit, 'timestamp', True)

    def handle_orders(self, client: Client, message):
        #
        #     {
        #         action: 'update',
        #         arg: {channel: 'orders'},
        #         data: [
        #           <same object in REST example>
        #         ]
        #     }
        #
        if self.orders is None:
            limit = self.safe_integer(self.options, 'ordersLimit', 1000)
            self.orders = ArrayCacheBySymbolById(limit)
        orders = self.orders
        arg = self.safe_dict(message, 'arg')
        channelName = self.safe_string(arg, 'channel')
        data = self.safe_list(message, 'data')
        for i in range(0, len(data)):
            order = self.parse_ws_order(data[i])
            symbol = order['symbol']
            messageHash = channelName + ':' + symbol
            orders.append(order)
            client.resolve(orders, messageHash)
            client.resolve(orders, channelName)

    def parse_ws_order(self, order, market: Market = None) -> Order:
        return self.parse_order(order, market)

    async def watch_positions(self, symbols: Strings = None, since: Int = None, limit: Int = None, params={}) -> List[Position]:
        """

        https://docs.blofin.com/index.html#ws-positions-channel

        watch all open positions
        :param str[]|None symbols: list of unified market symbols
        :param int [since]: the earliest time in ms to fetch positions for
        :param int [limit]: the maximum number of positions to retrieve
        :param dict params: extra parameters specific to the exchange API endpoint
        :returns dict[]: a list of `position structure <https://docs.ccxt.com/en/latest/manual.html#position-structure>`
        """
        await self.authenticate()
        await self.load_markets()
        newPositions = await self.watch_multiple_wrapper(False, 'positions', 'watchPositions', symbols, params)
        if self.newUpdates:
            return newPositions
        return self.filter_by_symbols_since_limit(self.positions, symbols, since, limit)

    def handle_positions(self, client: Client, message):
        #
        #     {
        #         arg: {channel: 'positions'},
        #         data: [
        #           <same object in REST example>
        #         ]
        #     }
        #
        if self.positions is None:
            self.positions = ArrayCacheBySymbolBySide()
        cache = self.positions
        arg = self.safe_dict(message, 'arg')
        channelName = self.safe_string(arg, 'channel')
        data = self.safe_list(message, 'data')
        newPositions = []
        for i in range(0, len(data)):
            position = self.parse_ws_position(data[i])
            newPositions.append(position)
            cache.append(position)
            messageHash = channelName + ':' + position['symbol']
            client.resolve(position, messageHash)

    def parse_ws_position(self, position, market: Market = None) -> Position:
        return self.parse_position(position, market)

    async def watch_multiple_wrapper(self, isPublic: bool, channelName: str, callerMethodName: str, symbolsArray: List[Any] = None, params={}):
        # underlier method for all watch-multiple symbols
        await self.load_markets()
        callerMethodName, params = self.handle_param_string(params, 'callerMethodName', callerMethodName)
        # if OHLCV method are being called, then symbols would be symbolsAndTimeframes(multi-dimensional) array
        isOHLCV = (channelName == 'candle')
        symbols = self.get_list_from_object_values(symbolsArray, 0) if isOHLCV else symbolsArray
        symbols = self.market_symbols(symbols, None, True, True)
        firstMarket = None
        firstSymbol = self.safe_string(symbols, 0)
        if firstSymbol is not None:
            firstMarket = self.market(firstSymbol)
        marketType = None
        marketType, params = self.handle_market_type_and_params(callerMethodName, firstMarket, params)
        if marketType != 'swap':
            raise NotSupported(self.id + ' ' + callerMethodName + '() does not support ' + marketType + ' markets yet')
        rawSubscriptions = []
        messageHashes = []
        if symbols is None:
            symbols = []
        symbolsLength = len(symbols)
        if symbolsLength > 0:
            for i in range(0, len(symbols)):
                current = symbols[i]
                market = None
                channel = channelName
                if isOHLCV:
                    market = self.market(current)
                    tfArray = symbolsArray[i]
                    tf = tfArray[1]
                    interval = self.safe_string(self.timeframes, tf, tf)
                    channel += interval
                else:
                    market = self.market(current)
                topic = {
                    'channel': channel,
                    'instId': market['id'],
                }
                rawSubscriptions.append(topic)
                messageHashes.append(channel + ':' + market['symbol'])
        else:
            rawSubscriptions.append({'channel': channelName})
            messageHashes.append(channelName)
        # private channel are difference, they only need plural channel name for multiple symbols
        if self.in_array(channelName, ['orders', 'positions']):
            rawSubscriptions = [{'channel': channelName}]
        request = self.get_subscription_request(rawSubscriptions)
        privateOrPublic = 'public' if isPublic else 'private'
        url = self.implode_hostname(self.urls['api']['ws'][marketType][privateOrPublic])
        return await self.watch_multiple(url, messageHashes, self.deep_extend(request, params), messageHashes)

    def get_subscription_request(self, args):
        return {
            'op': 'subscribe',
            'args': args,
        }

    def handle_message(self, client: Client, message):
        #
        # message examples
        #
        # {
        #   arg: {
        #     channel: "trades",
        #     instId: "DOGE-USDT",
        #   },
        #   event: "subscribe"
        # }
        #
        # incoming data updates' examples can be seen under each handler method
        #
        methods = {
            # public
            'pong': self.handle_pong,
            'trades': self.handle_trades,
            'books': self.handle_order_book,
            'tickers': self.handle_ticker,
            'candle': self.handle_ohlcv,  # candle1m, candle5m, etc
            # private
            'account': self.handle_balance,
            'orders': self.handle_orders,
            'positions': self.handle_positions,
        }
        method = None
        if message == 'pong':
            method = self.safe_value(methods, 'pong')
        else:
            event = self.safe_string(message, 'event')
            if event == 'subscribe':
                return
            elif event == 'login':
                future = self.safe_value(client.futures, 'authenticate_hash')
                future.resolve(True)
                return
            elif event == 'error':
                raise ExchangeError(self.id + ' error: ' + self.json(message))
            arg = self.safe_dict(message, 'arg')
            channelName = self.safe_string(arg, 'channel')
            method = self.safe_value(methods, channelName)
            if not method and channelName.find('candle') >= 0:
                method = methods['candle']
        if method:
            method(client, message)

    async def authenticate(self, params={}):
        self.check_required_credentials()
        milliseconds = self.milliseconds()
        messageHash = 'authenticate_hash'
        timestamp = str(milliseconds)
        nonce = 'n_' + timestamp
        auth = '/users/self/verify' + 'GET' + timestamp + '' + nonce
        signature = self.string_to_base64(self.hmac(self.encode(auth), self.encode(self.secret), hashlib.sha256))
        request = {
            'op': 'login',
            'args': [
                {
                    'apiKey': self.apiKey,
                    'passphrase': self.password,
                    'timestamp': timestamp,
                    'nonce': nonce,
                    'sign': signature,
                },
            ],
        }
        marketType = 'swap'  # for now
        url = self.implode_hostname(self.urls['api']['ws'][marketType]['private'])
        await self.watch(url, messageHash, self.deep_extend(request, params), messageHash)
